/**
 * Copyright (c) 2019 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Alexander Rose <alexander.rose@weirdbyte.de>
 */
import { __assign } from "tslib";
import { createComputeRenderable } from '../../renderable';
import { createComputeRenderItem } from '../../webgl/render-item';
import { TextureSpec, UniformSpec } from '../../renderable/schema';
import { ShaderCode } from '../../../mol-gl/shader-code';
import { ValueCell } from '../../../mol-util';
import { QuadSchema, QuadValues } from '../util';
import { getHistopyramidSum } from './sum';
import { isPowerOfTwo } from '../../../mol-math/misc';
import quad_vert from '../../../mol-gl/shader/quad.vert';
import reduction_frag from '../../../mol-gl/shader/histogram-pyramid/reduction.frag';
var HistopyramidReductionSchema = __assign(__assign({}, QuadSchema), { tPreviousLevel: TextureSpec('texture', 'rgba', 'float', 'nearest'), uSize: UniformSpec('f'), uTexSize: UniformSpec('f') });
var HistopyramidReductionRenderable;
function getHistopyramidReductionRenderable(ctx, initialTexture) {
    if (HistopyramidReductionRenderable) {
        ValueCell.update(HistopyramidReductionRenderable.values.tPreviousLevel, initialTexture);
        HistopyramidReductionRenderable.update();
        return HistopyramidReductionRenderable;
    }
    else {
        var values = __assign(__assign({}, QuadValues), { tPreviousLevel: ValueCell.create(initialTexture), uSize: ValueCell.create(0), uTexSize: ValueCell.create(0) });
        var schema = __assign({}, HistopyramidReductionSchema);
        var shaderCode = ShaderCode('reduction', quad_vert, reduction_frag);
        var renderItem = createComputeRenderItem(ctx, 'triangles', shaderCode, schema, values);
        HistopyramidReductionRenderable = createComputeRenderable(renderItem, values);
        return HistopyramidReductionRenderable;
    }
}
var LevelTexturesFramebuffers = [];
function getLevelTextureFramebuffer(ctx, level) {
    var textureFramebuffer = LevelTexturesFramebuffers[level];
    var size = Math.pow(2, level);
    if (textureFramebuffer === undefined) {
        var texture = ctx.resources.texture('image-float32', 'rgba', 'float', 'nearest');
        var framebuffer = ctx.resources.framebuffer();
        texture.attachFramebuffer(framebuffer, 0);
        textureFramebuffer = { texture: texture, framebuffer: framebuffer };
        textureFramebuffer.texture.define(size, size);
        LevelTexturesFramebuffers[level] = textureFramebuffer;
    }
    return textureFramebuffer;
}
function setRenderingDefaults(ctx) {
    var gl = ctx.gl, state = ctx.state;
    state.disable(gl.CULL_FACE);
    state.disable(gl.BLEND);
    state.disable(gl.DEPTH_TEST);
    state.disable(gl.SCISSOR_TEST);
    state.depthMask(false);
    state.colorMask(true, true, true, true);
    state.clearColor(0, 0, 0, 0);
}
export function createHistogramPyramid(ctx, inputTexture, scale) {
    var gl = ctx.gl, resources = ctx.resources;
    // printTexture(ctx, inputTexture, 2)
    if (inputTexture.getWidth() !== inputTexture.getHeight() || !isPowerOfTwo(inputTexture.getWidth())) {
        throw new Error('inputTexture must be of square power-of-two size');
    }
    // This part set the levels
    var levels = Math.ceil(Math.log(inputTexture.getWidth()) / Math.log(2));
    var maxSize = Math.pow(2, levels);
    // console.log('levels', levels, 'maxSize', maxSize)
    var pyramidTexture = resources.texture('image-float32', 'rgba', 'float', 'nearest');
    pyramidTexture.define(maxSize, maxSize);
    var framebuffer = resources.framebuffer();
    pyramidTexture.attachFramebuffer(framebuffer, 0);
    gl.clear(gl.COLOR_BUFFER_BIT);
    var levelTexturesFramebuffers = [];
    for (var i = 0; i < levels; ++i)
        levelTexturesFramebuffers.push(getLevelTextureFramebuffer(ctx, i));
    var renderable = getHistopyramidReductionRenderable(ctx, inputTexture);
    ctx.state.currentRenderItemId = -1;
    setRenderingDefaults(ctx);
    var offset = 0;
    for (var i = 0; i < levels; i++) {
        var currLevel = levels - 1 - i;
        var tf = levelTexturesFramebuffers[currLevel];
        tf.framebuffer.bind();
        // levelTextures[currLevel].attachFramebuffer(framebuffer, 0)
        var size = Math.pow(2, currLevel);
        // console.log('size', size, 'draw-level', currLevel, 'read-level', levels - i)
        gl.clear(gl.COLOR_BUFFER_BIT);
        gl.viewport(0, 0, size, size);
        ValueCell.update(renderable.values.uSize, Math.pow(2, i + 1) / maxSize);
        ValueCell.update(renderable.values.uTexSize, size);
        if (i > 0) {
            ValueCell.update(renderable.values.tPreviousLevel, levelTexturesFramebuffers[levels - i].texture);
            renderable.update();
        }
        ctx.state.currentRenderItemId = -1;
        renderable.render();
        pyramidTexture.bind(0);
        gl.copyTexSubImage2D(gl.TEXTURE_2D, 0, offset, 0, 0, 0, size, size);
        pyramidTexture.unbind(0);
        offset += size;
    }
    gl.finish();
    // printTexture(ctx, pyramidTexture, 2)
    //
    var finalCount = getHistopyramidSum(ctx, levelTexturesFramebuffers[0].texture);
    var height = Math.ceil(finalCount / Math.pow(2, levels));
    // const scale = Vec2.create(maxSize / inputTexture.width, maxSize / inputTexture.height)
    // console.log('height', height, 'finalCount', finalCount, 'scale', scale)
    return {
        pyramidTex: pyramidTexture,
        count: finalCount,
        height: height,
        levels: levels,
        scale: scale
    };
}
//# sourceMappingURL=reduction.js.map